"""A set of methods and classes for handing and creating sounds."""

import json

from math import log2

import aubio
import noisereduce as nr
import numpy as np

from moviepy.editor import VideoFileClip
from pydub import AudioSegment
from pydub.generators import Sine

from artbox.base import ArtBox

NOTES_FREQ = {
    "A": 440,
    "B": 494,
    "C": 523,
    "D": 587,
    "E": 659,
    "F": 698,
    "G": 784,
    "C0": 16.35,
    "C#0": 17.32,
    "D0": 18.35,
    "D#0": 19.45,
    "E0": 20.60,
    "F0": 21.83,
    "F#0": 23.12,
    "G0": 24.50,
    "G#0": 25.96,
    "A0": 27.50,
    "A#0": 29.14,
    "B0": 30.87,
    "C1": 32.70,
    "C#1": 34.65,
    "D1": 36.71,
    "D#1": 38.89,
    "E1": 41.20,
    "F1": 43.65,
    "F#1": 46.25,
    "G1": 49.00,
    "G#1": 51.91,
    "A1": 55.00,
    "A#1": 58.27,
    "B1": 61.74,
    "C2": 65.41,
    "C#2": 69.30,
    "D2": 73.42,
    "D#2": 77.78,
    "E2": 82.41,
    "F2": 87.31,
    "F#2": 92.50,
    "G2": 98.00,
    "G#2": 103.83,
    "A2": 110.00,
    "A#2": 116.54,
    "B2": 123.47,
    "C3": 130.81,
    "C#3": 138.59,
    "D3": 146.83,
    "D#3": 155.56,
    "E3": 164.81,
    "F3": 174.61,
    "F#3": 185.00,
    "G3": 196.00,
    "G#3": 207.65,
    "A3": 220.00,
    "A#3": 233.08,
    "B3": 246.94,
    "C4": 261.63,
    "C#4": 277.18,
    "D4": 293.66,
    "D#4": 311.13,
    "E4": 329.63,
    "F4": 349.23,
    "F#4": 369.99,
    "G4": 392.00,
    "G#4": 415.30,
    "A4": 440.00,
    "A#4": 466.16,
    "B4": 493.88,
    "C5": 523.25,
    "C#5": 554.37,
    "D5": 587.33,
    "D#5": 622.25,
    "E5": 659.26,
    "F5": 698.46,
    "F#5": 739.99,
    "G5": 783.99,
    "G#5": 830.61,
    "A5": 880.00,
    "A#5": 932.33,
    "B5": 987.77,
    "C6": 1046.50,
    "C#6": 1108.73,
    "D6": 1174.66,
    "D#6": 1244.51,
    "E6": 1318.51,
    "F6": 1396.91,
    "F#6": 1479.98,
    "G6": 1567.98,
    "G#6": 1661.22,
    "A6": 1760.00,
    "A#6": 1864.66,
    "B6": 1975.53,
    "C7": 2093.00,
    "C#7": 2217.46,
    "D7": 2349.32,
    "D#7": 2489.02,
    "E7": 2637.02,
    "F7": 2793.83,
    "F#7": 2959.96,
    "G7": 3135.96,
    "G#7": 3322.44,
    "A7": 3520.00,
    "A#7": 3729.31,
    "B7": 3951.07,
    "C8": 4186.01,
}


class Sound(ArtBox):
    """A set of methods for handing and creating sounds."""

    def process_notes(self, notes: list[str]) -> list[str]:
        """Process notes according to the available notes' table."""
        filtered_notes = []

        for note in notes:
            # example of n: C#1
            processed_note = ""
            while note:
                if note in NOTES_FREQ:
                    processed_note = note
                    break
                note = note[:-1]  # noqa: PLW2901

            if processed_note:
                filtered_notes.append(processed_note)
            elif filtered_notes:
                # replicate the last note
                filtered_notes.append(filtered_notes[-1])

        return filtered_notes

    def notes_to_audio(self):
        """
        Generate a simple melody using sine waves.

        Returns
        -------
        pydub.AudioSegment
            The generated melody.
        """
        notes_path = str(self.input_path)
        total_duration = float(self.args.get("duration", 0))

        if not total_duration:
            raise Exception("Argument `duration` was not given.")

        melody = AudioSegment.silent(0)

        notes = NOTES_FREQ

        # Define a simple sequence
        with open(notes_path, "r") as f:
            background_music_sequence = self.process_notes(json.load(f))

        note_duration = round(
            (total_duration / len(background_music_sequence)) * 1000
        )
        # Generate melody
        for note in background_music_sequence:
            tone = Sine(notes[note]).to_audio_segment(duration=note_duration)
            melody += tone

        melody.export(str(self.output_path), format="mp3")

    def convert_to_8bit_audio(self) -> None:
        """
        Extract audio from an MP4 file and convert it to a 16-bit.

        The result audio would be similar to the sound used by SNES and
        Sega Genesis.
        """
        video_path = str(self.input_path)
        output_path = str(self.output_path)

        # Load the video file
        video_clip = VideoFileClip(video_path)

        # Get audio as numpy array (stereo)
        audio_array = video_clip.audio.to_soundarray(
            fps=22050
        )  # Downsample to 22050 Hz

        # Convert to mono by averaging the two channels
        audio_mono = audio_array.mean(axis=1)

        # Reduce noise
        audio_denoised = nr.reduce_noise(y=audio_mono, sr=22050)

        # Convert to 8-bit PCM format
        audio_8bit = ((audio_denoised + 1) * 127.5).astype("uint8")

        # Create an AudioSegment from the 8-bit array
        audio_segment = AudioSegment(
            audio_8bit.tobytes(),
            frame_rate=22050,
            sample_width=1,  # 1 byte = 8 bits
            channels=1,
        )

        # Export as WAV file
        audio_segment.export(output_path, format="mp3")

        # Close the video clip
        video_clip.reader.close()

        print(
            "Audio has been extracted and converted to 8-bit format "
            f"with noise reduction. Output saved at '{output_path}'."
        )

    def frequency_to_note(self, frequency: float) -> str:
        """
        Convert a frequency in Hz to a musical note.

        Parameters
        ----------
        frequency : float
            Frequency in Hz.

        Returns
        -------
        str
            Corresponding musical note (e.g., "A4", "C#5").
        """
        notes = [
            "C",
            "C#",
            "D",
            "D#",
            "E",
            "F",
            "F#",
            "G",
            "G#",
            "A",
            "A#",
            "B",
        ]
        octave = int(frequency / 130.81)  # Based on C3 = 130.81 Hz
        n = int(round(12 * log2(frequency / 440.0)) + 49) % 12
        return f"{notes[n]}{octave}"

    def extract_notes_from_mp3(self) -> list:
        """Extra notes from a mp3 file."""
        mp3_path = str(self.input_path)
        output_notes = str(self.output_path)

        win_s = 2048  # Correct FFT size
        hop_s = win_s // 2  # Correct Hop size

        # Convert MP3 to WAV using pydub
        audio = AudioSegment.from_mp3(mp3_path)
        audio = audio.set_channels(1)  # Convert to mono
        samples = np.array(audio.get_array_of_samples())

        # Create pitch detection object
        pitch_o = aubio.pitch("default", win_s, hop_s, 44100)
        pitch_o.set_unit("Hz")

        notes = []
        for start in range(0, len(samples), hop_s):
            chunk = samples[start : start + hop_s]

            # Zero-pad the chunk if necessary
            if len(chunk) < hop_s:
                chunk = np.pad(
                    chunk,
                    (0, hop_s - len(chunk)),
                    "constant",
                    constant_values=0,
                )

            chunk = chunk.astype("float32")
            frequency = pitch_o(chunk)[0]

            # Convert frequency to musical note and add to the list
            if frequency > 0:  # Ignore zero frequencies (silence)
                note = self.frequency_to_note(frequency)
                notes.append(note)

        with open(output_notes, "w") as f:
            json.dump(notes, f)

        return notes



"""Base classes for ArtBox."""

from __future__ import annotations

from abc import ABC
from pathlib import Path


class ArtBox(ABC):
    """The base class for all ArtBox classes."""

    def __init__(self, args: dict[str, str]) -> None:
        """Initialize ArtBox class."""
        self.args: dict[str, str] = args
        self.input_path = Path(self.args.get("input-path", "/tmp"))
        self.output_path = Path(self.args.get("output-path", "/tmp"))



"""GGPEDIA Youtube."""
# mypy: disable-error-code="attr-defined"

from importlib import metadata as importlib_metadata


def get_version():
    """Return the program version."""
    try:
        return importlib_metadata.version(__name__)
    except importlib_metadata.PackageNotFoundError:  # pragma: no cover
        return "0.7.0"  # semantic-release


version = get_version()

__version__ = version
__author__ = "Ivan Ogasawara"
__email__ = "ivan.ogasawara@gmail.com"



"""
Set of tools for video handling.

ref: https://github.com/ethand91/python-youtube/blob/master/main.py
"""

from abc import abstractmethod

from moviepy.editor import AudioFileClip, VideoFileClip
from pytubefix import YouTube as PyYouTube

from artbox.base import ArtBox


class DownloadBase(ArtBox):
    """Set of tools for handing videos."""

    @abstractmethod
    def download(self):
        """Download a video."""
        ...


def _convert_srt_to_plain_text(srt_text: str) -> str:
    """
    Convert an SRT file to plain text by removing timestamps and formatting.

    Parameters
    ----------
    srt_file_path (str): Path to the SRT file.

    Returns
    -------
    str: The extracted plain text from the SRT file.
    """
    plain_text = []
    # Skip lines that are part of SRT formatting (timestamps, etc.)
    for line in srt_text.split("\n"):
        if line.strip() and not line.strip().isdigit() and "-->" not in line:
            plain_text.append(line.strip())

    return "\n".join(plain_text)


class Youtube(DownloadBase):
    """Set of tools for handing videos."""

    def download(self):
        """Download a youtube video."""
        resolution = self.args.get("resolution", "")
        video_url = self.args.get("url", "")

        if not video_url:
            raise Exception("Argument `url` not given.")

        video = PyYouTube(video_url)

        # Filter the stream by resolution if provided,
        # else get the highest resolution
        if resolution:
            video_stream = video.streams.filter(
                res=resolution, file_extension="mp4"
            ).first()
            if video_stream is None:
                raise Exception(
                    f"No video stream found with resolution {resolution}."
                )
        else:
            video_stream = video.streams.get_highest_resolution()

        try:
            video_stream.download(str(self.output_path))
            print("Video was downloaded successfully")
        except Exception as e:
            print(f"Failed to download video: {e}")

    def download_captions(self):
        """Download the English closed captions of a YouTube video."""
        video_url = self.args.get("url", "")
        lang = self.args.get("lang", "en")
        format = self.args.get("format", "text")

        yt = PyYouTube(video_url)
        caption = yt.captions.get_by_language_code(f"a.{lang}")

        if not caption:
            print(f"No captions found for language {lang}.")
            return

        # Save the captions to a file
        cc = caption.generate_srt_captions()
        with open(str(self.output_path), "w") as f:
            if format == "text":
                cc = _convert_srt_to_plain_text(cc)
            f.write(cc)
        print("Captions downloaded successfully.")


class Video(ArtBox):
    """Set of tools for handing videos."""

    def combine_video_and_audio(self) -> None:
        """
        Combine video and audio files to create a new MP4 file.

        The result will be clipped to the time of the shorter input
        (video or audio), and the audio will fade out smoothly over the last
        5 seconds.
        """
        video_path = self.args.get("video-path", "")
        audio_path = self.args.get("audio-path", "")
        output_path = str(self.output_path)

        if not video_path:
            raise Exception("Argument `video-path` not given.")

        if not audio_path:
            raise Exception("Argument `audio-path` not given.")

        # Load the video (without audio) from the MP4 file
        video_clip = VideoFileClip(video_path)
        video_clip = video_clip.without_audio()

        # Load the audio from the MP3 file
        audio_clip = AudioFileClip(audio_path)

        # Determine the shorter duration of the two clips
        min_duration = min(video_clip.duration, audio_clip.duration)

        # Clip both the video and audio to the shorter duration
        video_clip = video_clip.subclip(0, min_duration)
        audio_clip = audio_clip.subclip(0, min_duration)

        # Apply a 5-second fade-out effect to the audio
        audio_clip = audio_clip.audio_fadeout(5)

        # Set the audio of the video clip
        final_clip = video_clip.set_audio(audio_clip)

        # Write the result to the output file
        final_clip.write_videofile(output_path, codec="libx264")

        # Close all clips
        video_clip.close()
        audio_clip.close()
        final_clip.close()

    def extract_audio(self) -> None:
        """Extract audio from an MP4 file."""
        video_path = str(self.input_path)
        output_path = str(self.output_path)

        video_clip = VideoFileClip(video_path)
        audio_clip = video_clip.audio
        audio_clip.write_audiofile(output_path)
        audio_clip.close()
        video_clip.reader.close()

        print(f"Audio has been extracted. Output saved at '{output_path}'.")

    def remove_audio(self) -> None:
        """Remove the audio from an MP4 file."""
        # Load the video
        video = VideoFileClip(str(self.input_path))

        # Set the audio track to None
        video = video.set_audio(None)

        # Write the result to a file
        video.write_videofile(str(self.output_path), codec="libx264")



"""Cli functions to define the arguments and to call Makim."""

import typer

from typing_extensions import Annotated

from artbox import __version__
from artbox.sounds import Sound
from artbox.speech import SpeechFromText, SpeechToText
from artbox.videos import Video, Youtube

app = typer.Typer(
    name="Artbox",
    help="A set of tools for handling multimedia files.",
    epilog=(
        "If you have any problem, open an issue at: "
        "https://github.com/ggpedia/artbox"
    ),
)

app_sound = typer.Typer(
    name="sound",
    help="Audio processing commands for Artbox.",
    short_help="Audio processing commands.",
)
app_video = typer.Typer(
    name="video",
    help="Video processing commands for Artbox.",
    short_help="Video processing commands.",
)
app_speech = typer.Typer(
    name="speech",
    help="Speech processing commands for Artbox.",
    short_help="Speech processing commands.",
)
app_youtube = typer.Typer(
    name="youtube",
    help="YouTube processing commands for Artbox.",
    short_help="YouTube processing commands.",
)

app.add_typer(app_sound, name="sound")
app.add_typer(app_video, name="video")
app.add_typer(app_speech, name="speech")
app.add_typer(app_youtube, name="youtube")


@app.callback(invoke_without_command=True)
def main(
    ctx: typer.Context,
    version: bool = typer.Option(
        None,
        "--version",
        "-v",
        is_flag=True,
        help="Show the version and exit.",
    ),
) -> None:
    """Process commands for specific flags; otherwise, show the help menu."""
    if version:
        typer.echo(f"Version: {__version__}")
        raise typer.Exit()

    if ctx.invoked_subcommand is None:
        typer.echo(ctx.get_help())
        raise typer.Exit(0)


@app_speech.command("from-text")
def speech_from_text(
    title: Annotated[
        str, typer.Option("--title", help="Specify the name of the audio file")
    ] = "artbox",
    input_path: Annotated[
        str,
        typer.Option(
            "--input-path", help="Specify the path of the text file (txt)"
        ),
    ] = "",
    output_path: Annotated[
        str,
        typer.Option(
            "--output-path", help="Specify the path to store the audio file"
        ),
    ] = "",
    engine: Annotated[
        str,
        typer.Option(
            "--engine",
            help="Choose the text-to-speech engine (Options: edge-tts, gtts)",
        ),
    ] = "edge-tts",
    lang: Annotated[
        str,
        typer.Option(
            "--lang", help="Choose the language for audio generation"
        ),
    ] = "en",
    rate: Annotated[
        str,
        typer.Option("--rate", help="Decrease/Increase the rate level"),
    ] = "+0%",
    volume: Annotated[
        str,
        typer.Option("--volume", help="Decrease/Increase the volume level"),
    ] = "+0%",
    pitch: Annotated[
        str,
        typer.Option("--pitch", help="Decrease/Increase the pitch level"),
    ] = "+0Hz",
) -> None:
    """Convert text to speech."""
    args_dict = {
        "title": title,
        "input-path": input_path,
        "output-path": output_path,
        "engine": engine,
        "lang": lang,
        "rate": rate,
        "volume": volume,
        "pitch": pitch,
    }

    runner = SpeechFromText(args_dict)
    runner.convert()


@app_speech.command("to-text")
def speech_to_text(
    input_path: Annotated[
        str,
        typer.Option(
            "--input-path",
            help="Specify the path of the audio file (mp3 or wav)",
        ),
    ] = "",
    output_path: Annotated[
        str,
        typer.Option(
            "--output-path", help="Specify the path to store the text file"
        ),
    ] = "",
    engine: Annotated[
        str,
        typer.Option(
            "--engine",
            help="Choose the text-to-speech engine (Options: google)",
        ),
    ] = "google",
    lang: Annotated[
        str,
        typer.Option(
            "--lang", help="Choose the language for audio generation"
        ),
    ] = "en",
) -> None:
    """Convert text to speech."""
    args_dict = {
        "input-path": input_path,
        "output-path": output_path,
        "engine": engine,
        "lang": lang,
    }

    runner = SpeechToText(args_dict)
    runner.convert()


@app_sound.command("notes-to-audio")
def sound_notes_to_audio(
    input_path: Annotated[
        str,
        typer.Option(
            "--input-path", help="Specify the path of the input file"
        ),
    ] = "",
    output_path: Annotated[
        str,
        typer.Option(
            "--output-path", help="Specify the path to store the audio file"
        ),
    ] = "",
    duration: Annotated[
        str,
        typer.Option("--duration", help="Specify the duration of the audio"),
    ] = "",
) -> None:
    """Convert notes to audio."""
    args_dict = {
        "input-path": input_path,
        "output-path": output_path,
        "duration": duration,
    }

    runner = Sound(args_dict)
    runner.notes_to_audio()


@app_video.command("remove-audio")
def video_remove_audio(
    input_path: Annotated[
        str,
        typer.Option(
            "--input-path", help="Specify the path of the input video file"
        ),
    ] = "",
    output_path: Annotated[
        str,
        typer.Option(
            "--output-path", help="Specify the path to store the video file"
        ),
    ] = "",
) -> None:
    """Remove audio from video file."""
    args_dict = {
        "input-path": input_path,
        "output-path": output_path,
    }

    runner = Video(args_dict)
    runner.remove_audio()


@app_video.command("extract-audio")
def video_extract_audio(
    input_path: Annotated[
        str,
        typer.Option(
            "--input-path", help="Specify the path of the input video file"
        ),
    ] = "",
    output_path: Annotated[
        str,
        typer.Option(
            "--output-path",
            help="Specify the path to store the extracted audio file",
        ),
    ] = "",
) -> None:
    """Extract audio from video file."""
    args_dict = {
        "input-path": input_path,
        "output-path": output_path,
    }

    runner = Video(args_dict)
    runner.extract_audio()


@app_video.command("combine-video-and-audio")
def video_combine_audio_and_video(
    video_path: Annotated[
        str,
        typer.Option(
            "--video-path", help="Specify the path of the video file"
        ),
    ] = "",
    audio_path: Annotated[
        str,
        typer.Option(
            "--audio-path", help="Specify the path of the audio file"
        ),
    ] = "",
    output_path: Annotated[
        str,
        typer.Option(
            "--output-path",
            help="Specify the path to store the combined video file",
        ),
    ] = "",
) -> None:
    """Combine audio and video files."""
    args_dict = {
        "video-path": video_path,
        "audio-path": audio_path,
        "output-path": output_path,
    }

    runner = Video(args_dict)
    runner.combine_video_and_audio()


@app_youtube.command("download")
def youtube_download(
    url: Annotated[
        str,
        typer.Option(
            "--url", help="Specify the URL of the YouTube video to download"
        ),
    ] = "",
    output_path: Annotated[
        str,
        typer.Option(
            "--output-path",
            help="Specify the path to store the downloaded video file",
        ),
    ] = "",
    resolution: Annotated[
        str,
        typer.Option(
            "--resolution", help="Set the quality of the downloaded video"
        ),
    ] = "",
) -> None:
    """Download youtube video."""
    args_dict = {
        "url": url,
        "output-path": output_path,
        "resolution": resolution,
    }

    runner = Youtube(args_dict)
    runner.download()


@app_youtube.command("cc")
def youtube_cc(
    url: Annotated[
        str,
        typer.Option(
            "--url", help="Specify the URL of the YouTube video to download"
        ),
    ] = "",
    output_path: Annotated[
        str,
        typer.Option(
            "--output-path",
            help=(
                "Specify the path to store the downloaded video file "
                "(.srt, .txt)"
            ),
        ),
    ] = "/tmp/cc.txt",
    lang: Annotated[
        str,
        typer.Option("--lang", help="Set the CC language to be downloaded"),
    ] = "en",
    format: Annotated[
        str,
        typer.Option("--format", help="Set the CC format (srt, text)"),
    ] = "text",
) -> None:
    """Download youtube video CC."""
    args_dict = {
        "url": url,
        "output-path": output_path,
        "lang": lang,
        "format": format,
    }

    runner = Youtube(args_dict)
    runner.download_captions()



"""
Utilities for handling audio voices.

ref: https://thepythoncode.com/article/convert-text-to-speech-in-python
"""

import asyncio
import os
import random

from abc import ABC
from pathlib import Path

import edge_tts
import gtts
import speech_recognition as sr

from edge_tts import VoicesManager
from pydub import AudioSegment

from artbox.base import ArtBox


def convert_mp3_to_wav(input_path: str, output_path: str) -> None:
    """Convert from mp3 to wav."""
    sound = AudioSegment.from_mp3(input_path)
    sound.export(output_path, format="wav")


class Speech(ArtBox, ABC):
    """Set of methods for handing audio voices."""


class SpeechFromTextEngineBase(Speech):
    """Set of methods for handing audio voices."""

    def convert(self) -> None:
        """Convert text to audio speech."""
        ...


class SpeechFromText(Speech):
    """Speech class will run commands according to the selected engine."""

    engine: SpeechFromTextEngineBase

    def __init__(self, *args, **kwargs) -> None:
        """Initialize Speech class."""
        super().__init__(*args, **kwargs)
        engine = self.args.get("engine", "edge-tts")

        if engine == "edge-tts":
            self.engine: SpeechFromTextEngineBase = SpeechEngineMSEdgeTTS(
                *args, **kwargs
            )
        elif engine == "gtts":
            self.engine: SpeechFromTextEngineBase = SpeechEngineGTTS(
                *args, **kwargs
            )
        else:
            raise Exception(f"Engine {engine} not found.")

    def convert(self) -> None:
        """Convert text to audio speech."""
        return self.engine.convert()


class SpeechEngineGTTS(SpeechFromTextEngineBase):
    """Google-Text-To-Speech engine."""

    def convert(self) -> None:
        """Convert text to audio speech."""
        title: str = self.args.get("title", "")
        input_path: str = self.args.get("input-path", "")
        lang: str = self.args.get("lang", "en")

        if not title:
            raise Exception("Argument `title` not given")

        if not input_path:
            raise Exception("Argument `input_path` not given")

        with open(input_path, "r") as f:
            text = f.read()

        tts = gtts.gTTS(text, lang=lang, slow=False)
        tts.save(str(self.output_path))


class SpeechEngineMSEdgeTTS(SpeechFromTextEngineBase):
    """Microsoft Edge Text-To-Speech engine."""

    async def async_convert(self) -> None:
        """Convert text to audio speech in async mode."""
        title: str = self.args.get("title", "")
        input_path: str = self.args.get("input-path", "")
        lang: str = self.args.get("lang", "en")
        rate = self.args.get("rate", "+0%")
        volume = self.args.get("volume", "+0%")
        pitch = self.args.get("pitch", "+0Hz")

        if not title:
            raise Exception("Argument `title` not given")

        if not input_path:
            raise Exception("Argument `input_path` not given")

        with open(input_path, "r") as f:
            text = f.read()

        params = {"Locale": lang} if "-" in lang else {"Language": lang}
        voices = await VoicesManager.create()
        voice_options = voices.find(Gender="Female", **params)

        communicate = edge_tts.Communicate(
            text=text,
            voice=random.choice(voice_options)["Name"],
            rate=rate,
            volume=volume,
            pitch=pitch,
        )
        with open(self.output_path, "wb") as file:
            async for chunk in communicate.stream():
                if chunk["type"] == "audio":
                    file.write(chunk["data"])
                elif chunk["type"] == "WordBoundary":
                    print(f"WordBoundary: {chunk}")

    def convert(self) -> None:
        """Convert text to audio speech."""
        loop = asyncio.get_event_loop_policy().get_event_loop()
        try:
            loop.run_until_complete(self.async_convert())
        finally:
            loop.close()


class SpeechToText(Speech):
    """Speech to Text class."""

    def convert(self) -> None:
        """Recognize speech from MP# using various engines options."""
        file_path: str = str(self.input_path)

        if file_path.endswith("mp3"):
            self.convert_from_mp3()
            return

        if file_path.endswith("wav"):
            self.convert_from_wav()
            return

        raise Exception(
            "The file format is not valid. Valid types are mp3 and wav."
        )

    def convert_from_mp3(self) -> None:
        """Recognize speech from MP# using various engines options."""
        file_path: Path = self.input_path

        # Convert MP3 to WAV
        wav_path = str(file_path).replace(".mp3", ".wav")
        convert_mp3_to_wav(str(file_path), wav_path)

        self.input_path = Path(wav_path)
        self.convert_from_wav()

        # Cleanup: Remove the WAV file
        os.remove(wav_path)

    def convert_from_wav(self) -> None:
        """Recognize speech from WAVE using various engines options."""
        wav_path: str = str(self.input_path)
        output_path: str = str(self.output_path)
        language: str = self.args.get("lang", "en-US")
        engine: str = self.args.get("engine", "google")

        # Initialize recognizer
        recognizer = sr.Recognizer()

        with sr.AudioFile(wav_path) as source:
            audio_data = recognizer.record(source)
            kwargs = {"audio_data": audio_data, "language": language}
            try:
                if engine == "google":
                    text = recognizer.recognize_google(**kwargs)
                elif engine == "google_cloud":
                    text = recognizer.recognize_google_cloud(**kwargs)
                elif engine == "wit":
                    text = recognizer.recognize_wit(**kwargs)
                elif engine == "azure":
                    text = recognizer.recognize_azure(**kwargs)
                elif engine == "houndify":
                    text = recognizer.recognize_houndify(**kwargs)
                elif engine == "ibm":
                    text = recognizer.recognize_ibm(**kwargs)
                elif engine == "vosk":
                    text = recognizer.recognize_vosk(**kwargs)
                elif engine == "whisper":
                    text = recognizer.recognize_whisper(**kwargs)
                elif engine == "whisper-api":
                    text = recognizer.recognize_whisper_api(
                        kwargs.get("audio_data")
                    )
                else:
                    raise Exception(f"Engine '{engine}' is not supported.")
            except sr.UnknownValueError:
                raise Exception(
                    f"{engine.title()} could not understand the audio"
                )
            except sr.RequestError as e:
                raise Exception(
                    f"Could not request results from {engine.title()}; {e}"
                )

        with open(output_path, "w") as f:
            f.write(text)
